#include <stddef.h>

#include "Config.h"
#include "Converge.h"
#include "Memory.h"
#include "VM.h"
#include "Object.h"
#include "Class_Class.h"
#include "Object_Class.h"
#include "Dict.h"
#include "Func.h"
#include "String.h"
#include "List.h"
#include "Meta_C_Class.h"

#include "Modules/Exceptions.h"




void _Con_Class_Class_new(Con_VM* vm);
void _Con_Class_Class_new(Con_VM* vm);
void _Con_Class_Class_new_finish(Con_VM* vm);
void Con_Class_Class_new_populate(Con_VM* vm, Con_Value obj, Con_Value class);
void _Con_Class_Class_init(Con_VM* vm);
void _Con_Class_Class_to_str_func(Con_VM* vm);
void _Con_Class_Class_is_subclass_func(Con_VM* vm);



void Con_Class_Class_bootstrap(Con_VM* vm)
{
	Con_Value fields, new_func, init_func, to_str_func, supers, is_subclass_func;
	
	Con_Object_set_slot(vm, vm->builtins[CON_BUILTIN_CLASS_CLASS], "instance_of", vm->builtins[CON_BUILTIN_CLASS_CLASS]);
	Con_Object_set_slot(vm, vm->builtins[CON_BUILTIN_CLASS_CLASS], "name", Con_String_new_c_str(vm, "Class"));
	Con_Object_set_slot(vm, vm->builtins[CON_BUILTIN_CLASS_CLASS], "dcopy", Con_Dict_lookup(vm, Con_Object_get_slot(vm, vm->builtins[CON_BUILTIN_OBJECT_CLASS], "fields"), Con_String_new_c_str(vm, "dcopy")));
	
	fields = Con_Dict_new(vm);
	
	supers = Con_List_new(vm);
	Con_List_append(vm, supers, vm->builtins[CON_BUILTIN_OBJECT_CLASS]);
	Con_Object_set_slot(vm, vm->builtins[CON_BUILTIN_CLASS_CLASS], "supers", supers);
	
	init_func = Con_Func_new(vm, true, (Con_PC) (C_Function) _Con_Class_Class_init, PC_TYPE_C_FUNCTION, Con_String_new_c_str(vm, "init"), 0, NULL, NULL);
	Con_Object_set_slot(vm, vm->builtins[CON_BUILTIN_CLASS_CLASS], "init", init_func);

	Con_Object_set_slot(vm, vm->builtins[CON_BUILTIN_CLASS_CLASS], "has_slot", vm->builtins[CON_BUILTIN_DEFAULT_HAS_SLOT_FUNC]);

	Con_Object_set_slot(vm, vm->builtins[CON_BUILTIN_CLASS_CLASS], "get_slot", vm->builtins[CON_BUILTIN_DEFAULT_GET_SLOT_FUNC]);

	to_str_func = Con_Func_new(vm, true, (Con_PC) (C_Function) _Con_Class_Class_to_str_func, PC_TYPE_C_FUNCTION, Con_String_new_c_str(vm, "to_str"), 0, NULL, NULL);
	Con_Object_set_slot(vm, vm->builtins[CON_BUILTIN_CLASS_CLASS], "to_str", to_str_func);

	new_func = Con_Func_new(vm, true, (Con_PC) (C_Function) _Con_Class_Class_new, PC_TYPE_C_FUNCTION, Con_String_new_c_str(vm, "new"), 1, NULL, NULL);
	Con_Object_set_slot(vm, vm->builtins[CON_BUILTIN_OBJECT_CLASS], "new", new_func);
	Con_Object_set_slot(vm, vm->builtins[CON_BUILTIN_CLASS_CLASS], "new", new_func);

	// Notice that we have to place the is_subclass function into the Object class because that's
	// already been fully formed. There's no way around this hack.

	is_subclass_func = Con_Func_new(vm, true, (Con_PC) (C_Function) _Con_Class_Class_is_subclass_func, PC_TYPE_C_FUNCTION, Con_String_new_c_str(vm, "is_subclass"), 0, NULL, NULL);
	Con_Object_set_slot(vm, vm->builtins[CON_BUILTIN_CLASS_CLASS], "is_subclass", is_subclass_func);
	Con_Object_set_slot(vm, vm->builtins[CON_BUILTIN_OBJECT_CLASS], "is_subclass", is_subclass_func);
	
	Con_Dict_set_item(vm, fields, Con_String_new_c_str(vm, "new"), new_func);
	Con_Dict_set_item(vm, fields, Con_String_new_c_str(vm, "init"), init_func);
	Con_Dict_set_item(vm, fields, Con_String_new_c_str(vm, "to_str"), to_str_func);
	Con_Dict_set_item(vm, fields, Con_String_new_c_str(vm, "is_subclass"), is_subclass_func);
	Con_Object_set_slot(vm, vm->builtins[CON_BUILTIN_CLASS_CLASS], "fields", fields);
}



void _Con_Class_Class_new(Con_VM* vm)
{
	Con_Value self, var_args, new_obj, all_supers, supers, class;
	Con_Object* obj;
	Con_Meta_C_Class_Obj* meta_c_class = NULL;
	int i, j;

	Con_VM_decode_args(vm, "ov", &self, &var_args);
	
	all_supers = Con_List_new(vm);
	supers = Con_Object_get_slot(vm, self, "supers");
	i = 0;
	while (1) {
		for (j = 0; j < Con_List_get_size(vm, supers); j += 1) {
			class = Con_List_get_item(vm, supers, j);
			if (!Con_VM_is_type(vm, class, "c"))
				Con_Mod_Exceptions_quick(vm, "Type_Exception", 2, vm->builtins[CON_BUILTIN_CLASS_CLASS], Con_Object_get_slot(vm, class, "instance_of"));
			if (class.datum.object->type == CON_OBJECT_META_C_CLASS) {
				if (meta_c_class != NULL)
					Con_Mod_Exceptions_quick(vm, "Exception", 1, Con_String_new_c_str(vm, "Classes may inherit from a maximum of one builtin class."));
				meta_c_class = (Con_Meta_C_Class_Obj*) class.datum.object;
			}
			if (!Con_VM_is(vm, class, vm->builtins[CON_BUILTIN_OBJECT_CLASS]))
				Con_List_append(vm, all_supers, class);
		}
		i += 1;
		if (i >= Con_List_get_size(vm, all_supers))
			break;
		supers = Con_Object_get_slot(vm, Con_List_get_item(vm, all_supers, i), "supers");
	}

	if (meta_c_class != NULL) {
		obj = Con_malloc(vm, meta_c_class->obj_size, Con_MEMORY_OBJECT);
		Con_Object_init(vm, (Con_Object*) obj);
		obj->type = CON_NORMAL_OBJECT;
		new_obj.type = CON_VALUE_OBJECT;
		new_obj.datum.object = (Con_Object*) obj;
	}
	else
		new_obj = Con_Object_new(vm);

	Con_Object_set_slot(vm, new_obj, "instance_of", self);
	
	Con_Class_Class_new_populate(vm, new_obj, self);
	
	Con_VM_con_stack_push_value(vm, vm->continuation, new_obj);
	Con_VM_add_failure_frame(vm, (Con_PC) (C_Function) _Con_Class_Class_new_finish, PC_TYPE_C_FUNCTION);
	Con_VM_con_stack_push_value(vm, vm->continuation, new_obj);
	
	Con_VM_con_stack_push_value(vm, vm->continuation, Con_Object_get_slot(vm, new_obj, "init"));
	for (i = 0; i < Con_List_get_size(vm, var_args); i += 1) {
		Con_VM_con_stack_push_value(vm, vm->continuation, Con_List_get_item(vm, var_args, i));
	}
	
	Con_VM_apply(vm, Con_List_get_size(vm, var_args));
	Con_VM_con_stack_pop_value(vm, vm->continuation);

	Con_VM_return(vm);
}



void _Con_Class_Class_new_finish(Con_VM* vm)
{
	Con_VM_return(vm);
}



void Con_Class_Class_new_populate(Con_VM* vm, Con_Value new_obj, Con_Value class)
{
	char* slot_name;
	Con_Value fields, fields_keys, key, val, supers;
	int i;

	fields = Con_Object_get_slot(vm, class, "fields");
	fields_keys = Con_Dict_get_keys(vm, fields);
	for (i = 0; i < Con_List_get_size(vm, fields_keys); i += 1) {
		key = Con_List_get_item(vm, fields_keys, i);
		Con_VM_ensure_type(vm, key, "s", "fields contains keys which are not strings.");
		slot_name = ((Con_String_Object*) key.datum.object)->str;
		if (!Con_Object_has_slot(vm, new_obj, slot_name)) {
			val = Con_Dict_lookup(vm, fields, key);
			// We can't set container for things like ints...
			if (val.type == CON_VALUE_OBJECT)
				Con_Object_set_slot(vm, val, "container", class);
			Con_Object_set_slot_raw(vm, new_obj, slot_name, val);
		}
	}
	supers = Con_Object_get_slot(vm, class, "supers");
	for (i = 0; i < Con_List_get_size(vm, supers); i += 1) {
		Con_Class_Class_new_populate(vm, new_obj, Con_List_get_item(vm, supers, i));
	}
}



void _Con_Class_Class_init(Con_VM* vm)
{
	Con_Value self, name, supers, fields;

	Con_VM_decode_args(vm, "osld", &self, &name, &supers, &fields);

	Con_Object_set_slot(vm, self, "name", name);
	Con_Object_set_slot(vm, self, "supers", supers);
	Con_Object_set_slot(vm, self, "fields", fields);

	Con_VM_con_stack_push_value(vm, vm->continuation, vm->builtins[CON_BUILTIN_NULL_VAL]);

	Con_VM_return(vm);
}



void _Con_Class_Class_to_str_func(Con_VM* vm)
{
	Con_Value self, str;

	Con_VM_decode_args(vm, "o", &self);
	
	str = Con_String_new_c_str(vm, "<Class ");
	str = Con_String_add(vm, str, Con_Object_get_slot(vm, self, "name"));
	str = Con_String_add(vm, str, Con_String_new_c_str(vm, ">"));

	Con_VM_con_stack_push_value(vm, vm->continuation, str);

	Con_VM_return(vm);
}



void _Con_Class_Class_is_subclass_func(Con_VM* vm)
{
	Con_Value self, other, stack, candidate;
	int i;

	Con_VM_decode_args(vm, "cc", &self, &other);
	
	stack = Con_List_new(vm);
	Con_List_append(vm, stack, self);
	i = 0;
	while (i < Con_List_get_size(vm, stack)) {
		candidate = Con_List_get_item(vm, stack, i);
		if (Con_VM_is(vm, other, candidate)) {
			Con_VM_con_stack_push_value(vm, vm->continuation, vm->builtins[CON_BUILTIN_NULL_VAL]);
			Con_VM_return(vm);
		}
		Con_List_extend(vm, stack, Con_Object_get_slot(vm, candidate, "supers"));
		i += 1;
	}

	Con_VM_con_stack_push_value(vm, vm->continuation, vm->builtins[CON_BUILTIN_FAIL_VAL]);
	Con_VM_return(vm);
}
